import cv2 as cv
import os
from pathlib import Path
from random import randrange
import re
import time

class Ansciier:
    def __init__(self, params=None) -> None:
        if params is None:
            params = {}
        self.params = {}
        self.params.update(params)
        self.lowest_fps = 1000
        self.highest_fps = 0
        self.timer = 0
        self.frame_counter = 0

    @staticmethod
    def get_last_frame(path: Path, start_frame) -> int:
        if '{0}' not in str(path): return start_frame + 1

        basename = os.path.basename(path)

        # filter file based on basename path
        img_list = list(
            filter(
                lambda x: re.match(basename.format(r'\d+'), x),
                os.listdir(path.parent)))

        biggest_frame = 0
        print(img_list)
        for i in img_list:
            frame_number = int(re.findall(basename.format(r'(\d+)'), i)[0])
            if frame_number > biggest_frame: biggest_frame = frame_number

        return biggest_frame

    @staticmethod
    def colorize(text: str,
                 fgBGR = None,
                 bgBGR = None) -> str:
        # BGR is Blue, Green, Red
        colored_text = ''

        # colorize foreground (the text)
        if fgBGR is not None:
            fgB, fgG, fgR = fgBGR
            colored_text += f'\033[38;2;{fgR};{fgG};{fgB}m'

        # colorize background
        if bgBGR is not None:
            bgB, bgG, bgR = bgBGR
            colored_text += f'\033[48;2;{bgR};{bgG};{bgB}m'

        colored_text += f'{text}\x1b[0m' # close ansi sequence

        return colored_text

    def _debug(self, prev_time: int, curr_frame: int):
        delta_time = time.time() * 1000 - prev_time
        fps = 1000 / delta_time

        self.timer += delta_time
        self.frame_counter += 1

        if self.lowest_fps > fps: self.lowest_fps = fps
        if self.highest_fps < fps: self.highest_fps = fps

        print(f'frame: {curr_frame}')
        print(f'fps: {fps}')
        # print(delta_time)


    def _create_pixel(self, frame) -> str:
        tsize = os.get_terminal_size()

        aspect_ratio = self.params.get('aspect_ratio')
        dim = self.params.get('dimension')
        width = None
        height = tsize.lines - 3 if tsize.lines > 3 else tsize.lines

        if dim: # dimension
            width = dim[0]
            height = dim[1]
        elif aspect_ratio: # aspect ratio
            width = int(height * aspect_ratio[0] / aspect_ratio[1])
        else: # default terminal size
            width = int(height * frame.shape[1] / frame.shape[0])

        resized = cv.resize(frame, (width, height))
        pixel = '\033[H'
        for row in range(resized.shape[0]):
            for col in range(resized.shape[1]):
                if self.params.get('ascii'):
                    char = self.params.get('char')
                    if char:
                        pixel += self.colorize(char[0], resized[row, col])
                    else:
                        pixel += self.colorize(chr(randrange(33, 127)),
                                           resized[row, col])
                else:
                    pixel += self.colorize(' ', bgBGR=resized[row, col])
            pixel += '\n'
        return pixel

    def draw_from_img(self):
        prev_time = time.time() * 1000
        try:
            for i in range(self.params.get('start_frame'),
                        self.params.get('last_frame')):
                img_path = self.params.get('pic_path').__str__().format(i)

                if not os.path.exists(img_path):
                    continue

                img = cv.imread(img_path)
                ascii_pixel = self._create_pixel(img)

                
                delay = 1000 / self.params.get('max_fps')
                while time.time() * 1000 < prev_time + delay:
                    pass

                print(ascii_pixel, end='')

                # debugging purpose
                if self.params.get('verbose'):
                    self._debug(prev_time, i)
                    
                prev_time = time.time() * 1000
        except KeyboardInterrupt:
            return


    def draw_from_vid(self, camera: bool=False):
        cap = None
        if camera:
            cap = cv.VideoCapture(0)
        else:
            cap = cv.VideoCapture(self.params.get('pic_path').__str__())
            cap.set(1, self.params.get('start_frame'))

        prev_time = time.time() * 1000
        i = 1
        try:
            while cap.isOpened():
                ret, frame = cap.read()

                if not ret:
                    break

                ascii_pixel = self._create_pixel(frame)

                delay = 1000 / self.params.get('max_fps')
                while time.time() * 1000 < prev_time + delay:
                    pass

                print(ascii_pixel, end='') # print to terminal

                # debugging purpose
                if self.params.get('verbose'):
                    self._debug(prev_time, i)
                    
                prev_time = time.time() * 1000
                i += 1
        except KeyboardInterrupt:
            return
    
    def stats(self) -> None:
        print(f'\nTime taken: {round(self.timer / 1000, 2)} seconds')
        print(f'Average fps: {round(self.frame_counter / self.timer * 1000, 2)}')
        print(f'Highest fps: {round(self.highest_fps, 2)}')
        print(f'Lowest fps: {round(self.lowest_fps, 2)}')
        print(f'Frame printed: {self.frame_counter}')



import argparse

def parse_args():
    parser = argparse.ArgumentParser(
        description='Mimic image or video in your terminal')
    parser.add_argument(
        'path',
        type=str,
        help=('the path of the picture/video use "{0}" like "frame{0}.jpg" if '
              'the images name is frame1.jpg, frame2.jpg, ...'))
    parser.add_argument('--ascii',
                        help='whether using ascii characters or ansi escape code',
                        action='store_true')
    parser.add_argument(
        '--char',
        help=('pass a character to get fixed character instead of some '
              'random text (must be used it with "--ascii" opt)'),
        type=str,
        default=None)
    parser.add_argument('--fps', type=int,help='maximum fps', default=60)
    parser.add_argument('--start-frame',
                        type=int,
                        default=1,
                        metavar='NUMBER',
                        help='set the first frame played')
    parser.add_argument('--last-frame',
                        type=int,
                        default=None,
                        metavar='NUMBER',
                        help='set the last frame played (not work for video)')
    parser.add_argument('-v',
                        '--verbose',
                        help='show some information',
                        action='store_true')

    dim_group = parser.add_mutually_exclusive_group()
    dim_group.add_argument(
        '--dim',
        help='fixed dimension to display in terminal (example: 480x360)',
        metavar='WIDTHxHEIGHT',
        type=str,
        default=None)
    dim_group.add_argument(
        '--aspect-ratio',
        help=
        ('preferred aspect ratio to fix stretched image while printing in terminal, '
         'result will change dynamically based on terminal size (usage: 16:9, 4:3, 10:5, ...). '
         'Try increasing the height value if the image stretched vertically and '
         'vice versa if streched horizontally.'),
        metavar='RATIO',
        type=str,
        default=None)

    return parser.parse_args()



import mimetypes
import os
from pathlib import Path

from .Ansciier import Ansciier
from .options import parse_args

__version__ = '1.2.1'

def main():
    # will enable color on windows terminal
    if os.name == 'nt':
        os.system("")
        
    def exists(path: Path, start_frame: int) -> bool:
        if '{0}' not in str(path):
            return os.path.exists(path)
        
        return os.path.exists(str(path).format(start_frame))

    def truthy(res: str) -> bool:
        lower_res = res.lower()
        truth_list = ['yes',  'y', 'yup', 'ya', 'yeah']
        if lower_res in truth_list:
            return True
        else:
            return False
    args = parse_args()

    # constant
    PIC_PATH = Path(args.path).absolute()
    ASCII = args.ascii
    START_FRAME = args.start_frame
    LAST_FRAME = args.last_frame or Ansciier.get_last_frame(
        PIC_PATH, START_FRAME)
    ASPECT_RATIO = [int(x) for x in args.aspect_ratio.split(':', 1)
                    ] if args.aspect_ratio else None
    DIMENSION = [int(x) for x in args.dim.lower().split('x', 1)
                 ] if args.dim else None
    VERBOSE = args.verbose
    CHARACTER = args.char
    MAX_FPS = args.fps

    params = {
        'pic_path': PIC_PATH,
        'ascii': ASCII,
        'start_frame': START_FRAME,
        'last_frame': LAST_FRAME,
        'aspect_ratio': ASPECT_RATIO,
        'dimension': DIMENSION,
        'verbose': VERBOSE,
        'char': CHARACTER,
        'max_fps': MAX_FPS,
    }

    ansciier = Ansciier(params)

    if START_FRAME >= LAST_FRAME:
        print('Bad expression for start_frame and last_frame')
        return

    if os.path.basename(PIC_PATH).lower() == 'camera':
        ansciier.draw_from_vid(True)

    elif not exists(PIC_PATH, START_FRAME):
        print('File doesn\'t exist')

        if truthy(input('Do you want to use camera? [Y/n]')):
            ansciier.draw_from_vid(True)
        else:
            print('OK.')
            return

    elif 'image' in mimetypes.guess_type(PIC_PATH.__str__())[0]:
        ansciier.draw_from_img()
    elif 'video' in mimetypes.guess_type(PIC_PATH.__str__())[0]:
        ansciier.draw_from_vid(False)
    else:
        print('File is not supported')

    
    if VERBOSE:
        ansciier.stats()
    
    print('Done.')



from setuptools import setup, find_packages
from pathlib import Path
from src.ansciier import __version__

ROOT = Path(__file__).parent

with open(f'{ROOT}/README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

with open(f'{ROOT}/requirements.txt', 'r', encoding='utf-8') as f:
    reqs = [req.strip() for req in f.read().strip().split('\n')]

setup(
    name='ansciier',
    version=__version__,
    author='Cudiph',
    license='MIT',
    author_email='dwiaceromo@gmail.com',
    description='Command line program to mimic any image to your terminal',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Cudiph/ansciier',
    project_urls={
        'Bug Tracker': 'https://github.com/Cudiph/ansciier/issues',
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Environment :: Console',
        'Topic :: Multimedia',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Multimedia :: Graphics :: Graphics Conversion',
        'Topic :: Multimedia :: Graphics :: Presentation',
        'Topic :: Terminals',
    ],
    entry_points = {
        'console_scripts': ['ansciier = ansciier:main'],
    },
    package_dir={'': 'src'},
    install_requires=reqs,
    packages=find_packages(where='src'),
    python_requires='>=3.6',
)
